#include "paramSimul.h"
#include "const.h"
#include "evaluationConfigurerations.h"
#include "ue.h"
#include "sector.h"
#include "objectBSSectorUe.h"
#include "rvGenerater.h"
#include "paramFastFading.h"
#include "struct.h"
#include <iostream>
#include <Eigen/Dense>
#include <vector> 

void new_memory();
void Set_grid_size();
void Generate_TLSPs_RN_map();

void delete_memory();
void Set_auto_corr_matrix();
void Get_TLSPs_by_interpolation();
void Check_corr_var();

void Transform_TLSP_to_LSP();

double h_m(double);


VectorXd *****TLSPs_map; // [BS][floor][condition][x][y] (LSP)
VectorXd **corr_RN_vecter; //[UE][BS] (LSP)

MatrixXd **R; // [condition][LSP]
MatrixXd **inv_L_tilde;  // [condition][LSP]
RowVectorXd **lamda_n_T;  // [condition][LSP]

int size_x_grid, size_y_grid;
double max_x, min_x, max_y, min_y;


void Generate_LSP()
{
	
	Set_grid_size();
	new_memory();
	Set_auto_corr_matrix();
	cout << "Generate_TLSPs_RN_map..." << endl;
	Generate_TLSPs_RN_map();
	cout << "Get_TLSPs_by_interpolation..." << endl;
	Get_TLSPs_by_interpolation();

	//Check_corr_var();
	
	Transform_TLSP_to_LSP();
	delete_memory(); 
}

void Set_grid_size()
{
	if (test_environment<3)
	{
		size_x_grid = int(120./grid_interval+1) + 2;
		size_y_grid = int(50/ grid_interval+1) + 2;
	}
	else
	{
		size_x_grid = int(inter_site_distance * 8 / sqrt(3.) / grid_interval + 1) + 2;
		size_y_grid = int(inter_site_distance * 5 / grid_interval + 1) + 2; 
	}
}

void new_memory()
{
	TLSPs_map = new VectorXd ****[num_bs];
	
	for (int bs_idx = 0; bs_idx < num_bs; bs_idx++)
	{
		TLSPs_map[bs_idx] = new VectorXd ***[num_floor];

		for (int floor_idx = 0; floor_idx < num_floor; floor_idx++)
		{
			TLSPs_map[bs_idx][floor_idx] = new VectorXd**[num_propagation_condition];
			
			for (int idx = 0; idx < num_propagation_condition; idx++)
			{
				TLSPs_map[bs_idx][floor_idx][idx] = new VectorXd*[size_x_grid];

				for (int x = 0; x < size_x_grid; x++)
				{
					TLSPs_map[bs_idx][floor_idx][idx][x] = new VectorXd[size_y_grid];
					
					for (int y = 0; y < size_y_grid; y++)
					{
						if (idx == LOS)
						{
							TLSPs_map[bs_idx][floor_idx][idx][x][y] = VectorXd(7);
						}
						else
						{
							TLSPs_map[bs_idx][floor_idx][idx][x][y] = VectorXd(6);
						}
					}
				}
			}
		}
	}
	

	corr_RN_vecter = new VectorXd *[num_ue];

	for (int ue_idx = 0; ue_idx < num_ue; ue_idx++)
	{
		corr_RN_vecter[ue_idx] = new VectorXd [num_bs];

		for (int bs_idx = 0; bs_idx < num_bs; bs_idx++)
		{
			if (ue[ue_idx].propagation_condition[bs_idx] == LOS)
			{
				corr_RN_vecter[ue_idx][bs_idx] = VectorXd(7);
			}
			else
			{
				corr_RN_vecter[ue_idx][bs_idx] = VectorXd(6);
			}
		}
	}

	
	for (int ue_idx = 0; ue_idx < num_ue; ue_idx++)
	{
		for (int bs_idx = 0; bs_idx < num_bs; bs_idx++)
		{
			if(ue[ue_idx].propagation_condition[bs_idx] == LOS)
			{
				ue[ue_idx].TLSPs[bs_idx] = ArrayXd(7);
				ue[ue_idx].LSPs[bs_idx] = ArrayXd(7);
			}
			else
			{
				ue[ue_idx].TLSPs[bs_idx] = ArrayXd(6);
				ue[ue_idx].LSPs[bs_idx] = ArrayXd(6);
			}
			
		}
	}

	
	R = new MatrixXd*[num_propagation_condition];
	inv_L_tilde = new MatrixXd*[num_propagation_condition];
	lamda_n_T = new RowVectorXd*[num_propagation_condition];

	for (int idx = 0; idx < num_propagation_condition; idx++)
	{
		int num_LSP;

		if (idx == LOS) { num_LSP = 7; }
		else            { num_LSP = 6; }
		
		R[idx] = new MatrixXd[num_LSP];
		inv_L_tilde[idx] = new MatrixXd[num_LSP];
		lamda_n_T[idx] = new RowVectorXd[num_LSP];
		
		for (int LSP_idx = 0; LSP_idx < num_LSP; LSP_idx++)
		{
			R[idx][LSP_idx] = MatrixXd(num_neighbor+1, num_neighbor+1);
			inv_L_tilde[idx][LSP_idx] = MatrixXd(num_neighbor, num_neighbor);
		}
		
	}
	
}

void delete_memory()
{
	for (int bs_idx = 0; bs_idx < num_bs; bs_idx++)
	{
		for (int floor_idx = 0; floor_idx < num_floor; floor_idx++)
		{
			for (int idx = 0; idx < num_propagation_condition; idx++)
			{
				for (int x = 0; x < size_x_grid; x++)
				{
					delete[]TLSPs_map[bs_idx][floor_idx][idx][x];
				}
					
				delete[]TLSPs_map[bs_idx][floor_idx][idx];

			}
			delete[]TLSPs_map[bs_idx][floor_idx];
		}
		delete[]TLSPs_map[bs_idx];
	}
	delete[]TLSPs_map;


	for (int ue_idx = 0; ue_idx < num_ue; ue_idx++)
	{
		delete[]corr_RN_vecter[ue_idx];
	}

	delete[]corr_RN_vecter;


	for (int idx = 0; idx < num_propagation_condition; idx++)
	{
		delete[]R[idx];
		delete[]lamda_n_T[idx];
		delete[]inv_L_tilde[idx];
	}

	delete[]R;
	delete[]lamda_n_T;
	delete[]inv_L_tilde;
}

void Set_auto_corr_matrix()
{
	double alpha[7];
	int num_LSP;

	MatrixXd L = MatrixXd(num_neighbor+1, num_neighbor+1);
	MatrixXd L_tilde = MatrixXd(num_neighbor, num_neighbor);

	for (int propagation_condition = 0; propagation_condition < num_propagation_condition; propagation_condition++)
	{
		if (propagation_condition == LOS)	{ num_LSP = 7;}
		else  		    { num_LSP = 6;}

		if (propagation_condition == LOS)  // LOS case 7 Large-scale parameters else 6 LSP
		{
			for (int LSP_idx = 0; LSP_idx < num_LSP; LSP_idx++)
			{
				if (LSP_idx == 0) { alpha[0] = corr_dist_SF[propagation_condition]; }
				else if (LSP_idx == 1) { alpha[1] = corr_dist_K[propagation_condition]; }
				else if (LSP_idx == 2) { alpha[2] = corr_dist_DS[propagation_condition]; }
				else if (LSP_idx == 3) { alpha[3] = corr_dist_ASD[propagation_condition]; }
				else if (LSP_idx == 4) { alpha[4] = corr_dist_ASA[propagation_condition]; }
				else if (LSP_idx == 5) { alpha[5] = corr_dist_ZSD[propagation_condition]; }
				else if (LSP_idx == 6) { alpha[6] = corr_dist_ZSA[propagation_condition]; }
			}
		}
		else
		{
			for (int LSP_idx = 0; LSP_idx < num_LSP; LSP_idx++)
			{
				if (LSP_idx == 0) { alpha[0] = corr_dist_SF[propagation_condition]; }
				else if (LSP_idx == 1) { alpha[1] = corr_dist_DS[propagation_condition]; }
				else if (LSP_idx == 2) { alpha[2] = corr_dist_ASD[propagation_condition]; }
				else if (LSP_idx == 3) { alpha[3] = corr_dist_ASA[propagation_condition]; }
				else if (LSP_idx == 4) { alpha[4] = corr_dist_ZSD[propagation_condition]; }
				else if (LSP_idx == 5) { alpha[5] = corr_dist_ZSA[propagation_condition]; }
			}
		}
		
		
		for (int LSP_idx = 0; LSP_idx < num_LSP; LSP_idx++)
		{	
			double d = grid_interval / alpha[LSP_idx];

			if (num_neighbor == 4)
			{
				R[propagation_condition][LSP_idx] <<	1,				h_m(d),			h_m(2 * d),		h_m(d),			h_m(sqrt(2)*d),
														h_m(d),			1,				h_m(d),			h_m(sqrt(2)*d), h_m(d),
														h_m(2 * d),		h_m(d),			1,				h_m(sqrt(5)*d), h_m(sqrt(2)*d),
														h_m(d),			h_m(sqrt(2)*d), h_m(sqrt(5)*d), 1,				h_m(d),
														h_m(sqrt(2)*d), h_m(d),			h_m(sqrt(2)*d), h_m(d),			1;
			}
			else if (num_neighbor == 8)
			{
				R[propagation_condition][LSP_idx] <<	1,				h_m(d),			h_m(2 * d),		h_m(d),			 h_m(d),		  h_m(sqrt(5)*d),  h_m(d),			h_m(3 * d),		 h_m(sqrt(2)*d),
														h_m(d),			1,				h_m(d),			h_m(sqrt(2)*d),  h_m(sqrt(2)*d),  h_m(sqrt(2)*d),  h_m(2 * d),		h_m(2 * d),		 h_m(d),
														h_m(2 * d),		h_m(d),			1,				h_m(sqrt(5)*d),  h_m(sqrt(5)*d),  h_m(d),		   h_m(3 * d),		h_m(d),			 h_m(sqrt(2)*d),
														h_m(d),			h_m(sqrt(2)*d), h_m(sqrt(5)*d), 1,				 h_m(2 * d),	  h_m(sqrt(8)*d),  h_m(sqrt(2)*d),  h_m(sqrt(10)*d), h_m(d),
														h_m(d),			h_m(sqrt(2)*d), h_m(sqrt(5)*d), h_m(2 * d),		 1,				  h_m(2 * d),	   h_m(sqrt(2)*d),  h_m(sqrt(10)*d), h_m(sqrt(5)*d),
														h_m(sqrt(5)*d), h_m(sqrt(2)*d), h_m(d),			h_m(sqrt(8)*d),  h_m(2 * d),	  1,			   h_m(sqrt(10)*d), h_m(sqrt(2)*d),  h_m(sqrt(5)*d),
														h_m(d),			h_m(2 * d),		h_m(3 * d),		h_m(sqrt(2)*d),  h_m(sqrt(2)*d),  h_m(sqrt(10)*d), 1,				h_m(4 * d),		 h_m(sqrt(5)*d),
														h_m(3 * d),		h_m(2 * d),		h_m(d),			h_m(sqrt(10)*d), h_m(sqrt(10)*d), h_m(sqrt(2)*d),  h_m(4 * d),		1,				 h_m(sqrt(5)*d),
														h_m(sqrt(2)*d), h_m(d),			h_m(sqrt(2)*d), h_m(d),			 h_m(sqrt(5)*d),  h_m(sqrt(5)*d),  h_m(sqrt(5)*d),  h_m(sqrt(5)*d),  1;
			}
			else
			{
				cout << "check num_neighbor = 4 or 8" << endl;
			}

		}
		
		for (int LSP_idx = 0; LSP_idx < num_LSP; LSP_idx++)
		{
			LLT<MatrixXd> LLT_R(R[propagation_condition][LSP_idx]);
			L = LLT_R.matrixL();
			L_tilde = L.block(0, 0, num_neighbor, num_neighbor); //   L_tilde = LLT_R_tilde= L(size_row-1, size_col-1)



			inv_L_tilde[propagation_condition][LSP_idx] = L_tilde.inverse();
			lamda_n_T[propagation_condition][LSP_idx] = L.row(num_neighbor);
		}
	}

}

void Generate_TLSPs_RN_map()
{
	int num_LSP;
	MatrixXd iid_RN_map(size_x_grid, size_y_grid); // gaussian iid Random number grid
	
	VectorXd s_tilde(num_neighbor);
	VectorXd a(num_neighbor+1);
	LOCATION *relative_position_to_s_n;
	relative_position_to_s_n = new LOCATION[num_neighbor];
	
	vector <int> **available_neighbor;
	available_neighbor = new vector <int>*[size_x_grid];


	for (int x = 0; x<size_x_grid; x++)
	{
		available_neighbor[x] = new vector <int>[size_y_grid];
	}

	if (num_neighbor == 4)
	{
		// s3 sn
		// s0 s1 s2
		relative_position_to_s_n[0].x = -1;
		relative_position_to_s_n[0].y = -1;
		relative_position_to_s_n[1].x =  0;
		relative_position_to_s_n[1].y = -1;
		relative_position_to_s_n[2].x =  1;
		relative_position_to_s_n[2].y = -1;
		relative_position_to_s_n[3].x = -1;
		relative_position_to_s_n[3].y =  0;

	}
	else if (num_neighbor ==8)
	{	
		//
		//   s3 sn
		//s6 s0 s1 s2 s7
		//   s4    s5
		relative_position_to_s_n[0].x = -1;
		relative_position_to_s_n[0].y = -1;
		relative_position_to_s_n[1].x =  0;
		relative_position_to_s_n[1].y = -1;
		relative_position_to_s_n[2].x =  1;
		relative_position_to_s_n[2].y = -1;
		relative_position_to_s_n[3].x = -1;
		relative_position_to_s_n[3].y =  0;
		relative_position_to_s_n[4].x = -1;
		relative_position_to_s_n[4].y = -2;
		relative_position_to_s_n[5].x =  1;
		relative_position_to_s_n[5].y = -2;
		relative_position_to_s_n[6].x = -2;
		relative_position_to_s_n[6].y = -1;
		relative_position_to_s_n[7].x =  2;
		relative_position_to_s_n[7].y = -1;
	}
	else
	{
		cout << "check num_neighbor = 4 or 8" << endl;
	}

	for (int y = 0; y < size_y_grid; y++) 
	{
		for (int x = 0; x < size_x_grid; x++)
		{
			for (int neighbor_idx = 0; neighbor_idx < num_neighbor; neighbor_idx++)
			{
				LOCATION neighbor_location;

				neighbor_location.x = x + relative_position_to_s_n[neighbor_idx].x;
				neighbor_location.y = y + relative_position_to_s_n[neighbor_idx].y;
				if (0 <= neighbor_location.x  && neighbor_location.x < size_x_grid && 0 <= neighbor_location.y  && neighbor_location.y < size_y_grid)
				{
					available_neighbor[x][y].push_back(neighbor_idx);
				}
			}
			available_neighbor[x][y].push_back(num_neighbor); // self position
		}
	}

	
	for (int bs_idx = 0; bs_idx < num_bs; bs_idx++)
	{
		cout << "bs_idx : " << bs_idx <<"/"<< num_bs << endl;
	
		for (int floor_idx = 0; floor_idx < num_floor; floor_idx++)
		{
			for (int propagation_condition = 0; propagation_condition < num_propagation_condition; propagation_condition++)
			{
				if (propagation_condition == LOS)	{ num_LSP = 7; }
				else 								{ num_LSP = 6; }

				VectorXd **corr_RN_map;

				corr_RN_map = new VectorXd *[size_x_grid];

				for (int x=0; x< size_x_grid; x++)
				{
					corr_RN_map[x] = new VectorXd [size_y_grid];
					for (int y=0 ; y< size_y_grid; y++)
					{
						corr_RN_map[x][y] = VectorXd(num_LSP);
					}
				}
				
				
				for (int LSP_idx = 0; LSP_idx < num_LSP; LSP_idx++)
				{
					
					RowVectorXd _lamda_n_T(num_neighbor + 1);
					double a_n;
					double s_n;
					_lamda_n_T << lamda_n_T[propagation_condition][LSP_idx];


					for (int x = 0; x < size_x_grid; x++)
					{
						for (int y = 0; y < size_y_grid; y++)
						{
							iid_RN_map(x, y) = Rand.n();
						}
					}
										

					for (int y = 0; y < size_y_grid; y++)
					{

						for (int x = 0; x < size_x_grid; x++)
						{
							a_n = iid_RN_map(x, y);

							if (available_neighbor[x][y].size() == num_neighbor + 1)
							{
								for (int s_idx = 0; s_idx < num_neighbor; s_idx++)
								{
									int neighbor_pos_x = x + int(relative_position_to_s_n[s_idx].x);
									int neighbor_pos_y = y + int(relative_position_to_s_n[s_idx].y);

									s_tilde(s_idx) = corr_RN_map[neighbor_pos_x][neighbor_pos_y](LSP_idx);
								}

								a << inv_L_tilde[propagation_condition][LSP_idx] * s_tilde, a_n;

								s_n = _lamda_n_T * a;
							}
							else
							{
								int num_available_neighbor = available_neighbor[x][y].size() - 1;

								MatrixXd _R(num_available_neighbor + 1, num_available_neighbor + 1);
								MatrixXd _L(num_available_neighbor + 1, num_available_neighbor + 1);
								MatrixXd _L_tilde(num_available_neighbor, num_available_neighbor);
								MatrixXd _inv_L_tilde(num_available_neighbor, num_available_neighbor);
								RowVectorXd _lamda_n_T(num_available_neighbor + 1);
								VectorXd _a(num_available_neighbor + 1);
								VectorXd _s_tilde(num_available_neighbor);


								if (num_available_neighbor > 0)
								{

									for (int available_neighbor_idx = 0; available_neighbor_idx < num_available_neighbor; available_neighbor_idx++)
									{
										int neighbor_pos_x = x + int(relative_position_to_s_n[available_neighbor[x][y][available_neighbor_idx]].x);
										int neighbor_pos_y = y + int(relative_position_to_s_n[available_neighbor[x][y][available_neighbor_idx]].y);

										_s_tilde(available_neighbor_idx) = corr_RN_map[neighbor_pos_x][neighbor_pos_y](LSP_idx);

									}
								}

								for (int R_prime_row_idx = 0; R_prime_row_idx < num_available_neighbor + 1; R_prime_row_idx++)
								{
									for (int R_prime_column_idx = 0; R_prime_column_idx < num_available_neighbor + 1; R_prime_column_idx++)
									{
										_R(R_prime_row_idx, R_prime_column_idx) = R[propagation_condition][LSP_idx](available_neighbor[x][y][R_prime_row_idx], available_neighbor[x][y][R_prime_column_idx]);
									}

								}

								if (num_available_neighbor > 0)
								{
									LLT<MatrixXd> LLT_R(_R);
									_L = LLT_R.matrixL();
									_L_tilde = _L.block(0, 0, num_available_neighbor, num_available_neighbor); //   L_tilde = LLT_R_tilde= L(size_row-1, size_col-1)
									_inv_L_tilde = _L_tilde.inverse();
									_lamda_n_T = _L.row(num_available_neighbor);
									_a << _inv_L_tilde * _s_tilde, a_n;
									s_n = _lamda_n_T * _a;
								}
								else
								{
									s_n = a_n;
								}
							}

							corr_RN_map[x][y](LSP_idx) = s_n;
						}
					}
					
				}


				for (int x=0; x< size_x_grid; x++)
				{
					for (int y =0 ; y< size_y_grid; y++)
					{
						switch (propagation_condition)
						{
						case LOS:
							TLSPs_map[bs_idx][floor_idx][propagation_condition][x][y] = sqrt_corr_matrx_LOS * corr_RN_map[x][y];
							break;
						case NLOS:
							TLSPs_map[bs_idx][floor_idx][propagation_condition][x][y] = sqrt_corr_matrx_NLOS* corr_RN_map[x][y];
							break;
						case OUT2IN:
							TLSPs_map[bs_idx][floor_idx][propagation_condition][x][y] = sqrt_corr_matrx_OUT2IN* corr_RN_map[x][y];
							break;
						default:
							cout << "check propagation_condition " << endl;
						}
						/*
						cout << " --------------------------------------------------------" << endl;
						cout << TLSPs_map[bs_idx][floor_idx][propagation_condition][x][y] << endl;
						cout << endl;
						cout << sqrt_corr_matrx_LOS << endl;
						cout << endl;
						cout << corr_RN_map[x][y] << endl;
						cout << endl;
						getchar();
						*/
						
					}
					
				}

				
				delete[]corr_RN_map;
			}
		}
	}


	for (int x = 0; x<size_x_grid; x++)
	{
		delete []available_neighbor[x];
	}
	delete[] available_neighbor;
	delete[] relative_position_to_s_n;

}
/*===================================================================
FUNCTION: Get_TLSPs_by_interpolation()

AUTHOR: G.J

DESCRIPTION: 
 - 

NOTES: 

===================================================================*/

void Get_TLSPs_by_interpolation()
{
	int x_[4];
	int y_[4];
	double x_pos;
	double y_pos;
	double D_corr = grid_interval;
	
	for (int bs_idx = 0; bs_idx < num_bs; bs_idx++)
	{
		for (int ue_idx = 0; ue_idx < num_ue; ue_idx++)
		{
			ArrayXd TLSPs[4];
			int propagation_condition = ue[ue_idx].propagation_condition[bs_idx];
			int floor_idx = ue[ue_idx].floor_idx;

			if (propagation_condition == LOS)
			{
				for (int idx=0; idx<4 ; idx++)
				{
					TLSPs[idx] = ArrayXd(7);
				}
				
			}
			else
			{
				for (int idx = 0; idx<4; idx++)
				{
					TLSPs[idx] = ArrayXd(6);
				}
			}

			
			if (test_environment < 3)
			{
				x_[0] = int((ue[ue_idx].pos_indoor.x + floor(size_x_grid / 2)*D_corr) / D_corr);
				y_[0] = int((ue[ue_idx].pos_indoor.y + floor(size_y_grid / 2)*D_corr) / D_corr) + 1;

				x_pos = (ue[ue_idx].pos_indoor.x + floor(size_x_grid / 2)*D_corr) - x_[0] * D_corr;
				y_pos = (ue[ue_idx].pos_indoor.y + floor(size_y_grid / 2)*D_corr) - (y_[0]-1) * D_corr;
			}
			else
			{
				x_[0] = int((ue[ue_idx].pos[bs_idx].x + floor(size_x_grid / 2)*D_corr) / D_corr);
				y_[0] = int((ue[ue_idx].pos[bs_idx].y + floor(size_y_grid / 2)*D_corr) / D_corr) + 1;
				x_pos = (ue[ue_idx].pos[bs_idx].x + floor(size_x_grid / 2)*D_corr) - x_[0] * D_corr;
				y_pos = (ue[ue_idx].pos[bs_idx].y + floor(size_y_grid / 2)*D_corr) - (y_[0] - 1) * D_corr;
			}
			
			x_[1] = x_[0] + 1;
			y_[1] = y_[0];
			x_[2] = x_[0] + 1;
			y_[2] = y_[0] - 1;
			x_[3] = x_[0];
			y_[3] = y_[0] - 1;
			
			TLSPs[0] << TLSPs_map[bs_idx][floor_idx][propagation_condition][x_[0]][y_[0]].array();
			TLSPs[1] << TLSPs_map[bs_idx][floor_idx][propagation_condition][x_[1]][y_[1]].array();
			TLSPs[2] << TLSPs_map[bs_idx][floor_idx][propagation_condition][x_[2]][y_[2]].array();
			TLSPs[3] << TLSPs_map[bs_idx][floor_idx][propagation_condition][x_[3]][y_[3]].array();

			

			ue[ue_idx].TLSPs[bs_idx] = sqrt(1 - x_pos / D_corr)*( TLSPs[0] * sqrt(y_pos / D_corr) + TLSPs[3] * sqrt(1 - y_pos / D_corr) ) + ( TLSPs[1] * sqrt(y_pos / D_corr) + TLSPs[2] * sqrt(1 - y_pos / D_corr) )*sqrt(x_pos / D_corr);
			//ue[ue_idx].TLSPs[bs_idx] = (1 - x_pos / D_corr)*(TLSPs[0] * (y_pos / D_corr) + TLSPs[3] * (1 - y_pos / D_corr)) + (TLSPs[1] * (y_pos / D_corr) + TLSPs[2] * (1 - y_pos / D_corr))*(x_pos / D_corr);
		}
	}
	
}

void Transform_TLSP_to_LSP()
{
	double fc = carrier_freq / 1000000000.;

	if (calibration == 1)
	{
		for (int bs_idx = 0; bs_idx < num_bs; bs_idx++)
		{
			for (int ue_idx = 0; ue_idx < num_ue; ue_idx++)
			{
				for (int i = 0; i < ue[ue_idx].TLSPs[bs_idx].size(); i++)
				{
					ue[ue_idx].TLSPs[bs_idx](i) = 1.;
				}
			}
		}
	}


	for (int bs_idx = 0; bs_idx < num_bs; bs_idx++)
	{
		for (int ue_idx = 0; ue_idx < num_ue; ue_idx++)
		{
			ue[ue_idx].LSPs[bs_idx](0) = ue[ue_idx].sigma_SF[bs_idx] * ue[ue_idx].TLSPs[bs_idx](0); // Generate SF[dB]

			if (ue[ue_idx].propagation_condition[bs_idx] == LOS)
			{
				
				ue[ue_idx].LSPs[bs_idx](LOS_K)   = K_std * ue[ue_idx].TLSPs[bs_idx](LOS_K) + K_mean;// [dB]
				ue[ue_idx].LSPs[bs_idx](LOS_DS)  = pow(10, log_DS_std[LOS]  * ue[ue_idx].TLSPs[bs_idx](LOS_DS)  + log_DS_mean[LOS] );					
				ue[ue_idx].LSPs[bs_idx](LOS_ASD) = MIN(pow(10., log_ASD_std[LOS] * ue[ue_idx].TLSPs[bs_idx](LOS_ASD) + log_ASD_mean[LOS]),104.);
				ue[ue_idx].LSPs[bs_idx](LOS_ASA) = MIN(pow(10., log_ASA_std[LOS] * ue[ue_idx].TLSPs[bs_idx](LOS_ASA) + log_ASA_mean[LOS]),104.);
				ue[ue_idx].LSPs[bs_idx](LOS_ZSA) = MIN(pow(10., log_ZSA_std[LOS] * ue[ue_idx].TLSPs[bs_idx](LOS_ZSA) + log_ZSA_mean[LOS]),52.);
			}
			else if (ue[ue_idx].propagation_condition[bs_idx] == NLOS)
			{

				ue[ue_idx].LSPs[bs_idx](NLOS_DS)  = pow(10., log_DS_std[NLOS]  * ue[ue_idx].TLSPs[bs_idx](NLOS_DS)  + log_DS_mean[NLOS]);		
				ue[ue_idx].LSPs[bs_idx](NLOS_ASD) = MIN(pow(10., log_ASD_std[NLOS] * ue[ue_idx].TLSPs[bs_idx](NLOS_ASD) + log_ASD_mean[NLOS]),104.);
				ue[ue_idx].LSPs[bs_idx](NLOS_ASA) = MIN(pow(10., log_ASA_std[NLOS] * ue[ue_idx].TLSPs[bs_idx](NLOS_ASA) + log_ASA_mean[NLOS]),104.);
				ue[ue_idx].LSPs[bs_idx](NLOS_ZSA) = MIN(pow(10., log_ZSA_std[NLOS] * ue[ue_idx].TLSPs[bs_idx](NLOS_ZSA) + log_ZSA_mean[NLOS]),52.);
			}
			else if (ue[ue_idx].propagation_condition[bs_idx] == OUT2IN)
			{
				ue[ue_idx].LSPs[bs_idx](OUT2IN_DS)  = pow(10., log_DS_std[OUT2IN]  * ue[ue_idx].TLSPs[bs_idx](OUT2IN_DS)  + log_DS_mean[OUT2IN]);
				ue[ue_idx].LSPs[bs_idx](OUT2IN_ASD) = MIN(pow(10., log_ASD_std[OUT2IN] * ue[ue_idx].TLSPs[bs_idx](OUT2IN_ASD) + log_ASD_mean[OUT2IN]),104.);
				ue[ue_idx].LSPs[bs_idx](OUT2IN_ASA) = MIN(pow(10., log_ASA_std[OUT2IN] * ue[ue_idx].TLSPs[bs_idx](OUT2IN_ASA) + log_ASA_mean[OUT2IN]),104.);
				ue[ue_idx].LSPs[bs_idx](OUT2IN_ZSA) = MIN(pow(10., log_ZSA_std[OUT2IN] * ue[ue_idx].TLSPs[bs_idx](OUT2IN_ZSA) + log_ZSA_mean[OUT2IN]),52.);
			
			}
			else
			{
				cout << "check Transform_TLSP_to_LSP" << endl;
				getchar();
			}
			
		}
	}

	for (int bs_idx = 0; bs_idx < num_bs; bs_idx++)
	{
		for (int ue_idx = 0; ue_idx < num_ue; ue_idx++)
		{
			double d_2D = ue[ue_idx]._2D_distance[bs_idx];
			double h_UT;

			if (calibration == true)
			{
				h_UT = 1.5;// ue[ue_idx].h_UT;
			}
			else
			{
				h_UT =  ue[ue_idx].h_UT;
			}

			if (pathloss_model == InH_A || pathloss_model == InH_B)
			{
				if (ue[ue_idx].propagation_condition[bs_idx] == LOS)
				{
					ue[ue_idx].LSPs[bs_idx](LOS_ZSD) = MIN(pow(10., log_ZSD_std[LOS] * ue[ue_idx].TLSPs[bs_idx](LOS_ZSD) + log_ZSD_mean[LOS]),52.);
					
				}
				else
				{
					ue[ue_idx].LSPs[bs_idx](NLOS_ZSD) = MIN(pow(10., log_ZSD_std[NLOS] * ue[ue_idx].TLSPs[bs_idx](NLOS_ZSD) + log_ZSD_mean[NLOS]),52.);
				}
				ue[ue_idx].ZOD_offset[bs_idx] =0;
				
			}
			else if (pathloss_model == UMa_A)
			{
				if ((0.5 < fc) && (fc < 6.))
				{
					log_ZSD_mean[LOS]  = MAX(-0.5, -2.1*(d_2D / 1000.) - 0.01*(h_UT - 1.5) + 0.75);
					log_ZSD_mean[NLOS] = MAX(-0.5, -2.1*(d_2D / 1000.) - 0.01*(h_UT - 1.5) + 0.9);
					log_ZSD_std[LOS] = 0.4;
					log_ZSD_std[NLOS] = 0.49;
					ZOD_offset[LOS] = 0;
					ZOD_offset[NLOS] = -pow(10., -0.62*log10(MAX(10., d_2D)) + 1.93 - 0.07*(h_UT - 1.5));
					
				}
				else if ((6. < fc) && (fc <= 100.))
				{
					log_ZSD_mean[LOS]  = MAX(-0.5, -2.1*(d_2D / 1000.) - 0.01*(h_UT - 1.5) + 0.75);
					log_ZSD_mean[NLOS] = MAX(-0.5, -2.1*(d_2D / 1000.) - 0.01*(h_UT - 1.5) + 0.9);
					log_ZSD_std[LOS] = 0.4;
					log_ZSD_std[NLOS] = 0.49;
					ZOD_offset[LOS] = 0;
					ZOD_offset[NLOS] = 7.66*log10(fc) - 5.96 - pow(10, (0.208*log10(fc) - 0.782)*log10(MAX(25., d_2D)) - 0.13*log10(fc) + 2.03);

				}

				ue[ue_idx].ZOD_offset[bs_idx] = ZOD_offset[ue[ue_idx].propagation_condition[bs_idx]];

				if (ue[ue_idx].propagation_condition[bs_idx] == LOS)
				{
					ue[ue_idx].LSPs[bs_idx](LOS_ZSD) = MIN(pow(10., log_ZSD_std[LOS] * ue[ue_idx].TLSPs[bs_idx](LOS_ZSD) + log_ZSD_mean[LOS]),52.);
				}
				else
				{
					ue[ue_idx].LSPs[bs_idx](NLOS_ZSD) = MIN(pow(10., log_ZSD_std[NLOS] * ue[ue_idx].TLSPs[bs_idx](NLOS_ZSD) + log_ZSD_mean[NLOS]),52.);
				}
			}
			else if (pathloss_model == UMa_B)
			{
				log_ZSD_mean[LOS] = MAX(-0.5, -2.1*(d_2D / 1000.) - 0.01*(h_UT - 1.5) + 0.75);
				log_ZSD_mean[NLOS] = MAX(-0.5, -2.1*(d_2D / 1000.) - 0.01*(h_UT - 1.5) + 0.9);
				log_ZSD_std[LOS] = 0.4;
				log_ZSD_std[NLOS] = 0.49;
				ZOD_offset[LOS] = 0;
				ZOD_offset[NLOS] = 7.66*log10(fc) - 5.96 - pow(10., (0.208*log10(fc) - 0.782)*log10(MAX(25, d_2D)) - 0.13*log10(fc) + 2.03 - 0.07*(h_UT - 1.5));

				ue[ue_idx].ZOD_offset[bs_idx] = ZOD_offset[ue[ue_idx].propagation_condition[bs_idx]];

				if (ue[ue_idx].propagation_condition[bs_idx] == LOS)
				{
					ue[ue_idx].LSPs[bs_idx](LOS_ZSD) = MIN(pow(10., log_ZSD_std[LOS] * ue[ue_idx].TLSPs[bs_idx](LOS_ZSD) + log_ZSD_mean[LOS]),52.);
				}
				else
				{
					ue[ue_idx].LSPs[bs_idx](NLOS_ZSD) = MIN(pow(10., log_ZSD_std[NLOS] * ue[ue_idx].TLSPs[bs_idx](NLOS_ZSD) + log_ZSD_mean[NLOS]),52.);
				}
			}
			else if (pathloss_model == UMi_A)
			{
				if ((0.5 < fc) && (fc < 6.))
				{
					log_ZSD_mean[LOS] = MAX(-0.5, -2.1*(d_2D / 1000.) + 0.01*abs(h_UT - h_BS) + 0.75);
					log_ZSD_mean[NLOS] = MAX(-0.5, -2.1*(d_2D / 1000.) + 0.01*MAX((h_UT - h_BS),0) + 0.9);
					log_ZSD_std[LOS] = 0.4;
					log_ZSD_std[NLOS] = 0.6;
					ZOD_offset[LOS] = 0;
					ZOD_offset[NLOS] = -pow(10, -0.55*log10(MAX(10., d_2D)) + 1.6);
				}
				else if ((6 < fc) && (fc <= 100))
				{
					log_ZSD_mean[LOS] = MAX(-0.21, -14.8*(d_2D / 1000) + 0.01*abs(h_UT - h_BS) + 0.83);
					log_ZSD_mean[NLOS] = MAX(-0.5, -3.1*(d_2D / 1000) + 0.01*MAX((h_UT - h_BS), 0) + 0.2);
					log_ZSD_std[LOS] = 0.35;
					log_ZSD_std[NLOS] = 0.35;
					ZOD_offset[LOS] = 0;
					ZOD_offset[NLOS] = -pow(10., -1.5*log10(MAX(10., d_2D)) + 3.3);

				}

				ue[ue_idx].ZOD_offset[bs_idx] = ZOD_offset[ue[ue_idx].propagation_condition[bs_idx]];

				if (ue[ue_idx].propagation_condition[bs_idx] == LOS)
				{
					ue[ue_idx].LSPs[bs_idx](LOS_ZSD) = MIN(pow(10, log_ZSD_std[LOS] * ue[ue_idx].TLSPs[bs_idx](LOS_ZSD) + log_ZSD_mean[LOS]),52);
				}
				else
				{
					ue[ue_idx].LSPs[bs_idx](NLOS_ZSD) = MIN(pow(10, log_ZSD_std[NLOS] * ue[ue_idx].TLSPs[bs_idx](NLOS_ZSD) + log_ZSD_mean[NLOS]),52);
				}
			}
			else if (pathloss_model == UMi_B)
			{
				log_ZSD_mean[LOS] = MAX(-0.21, -14.8*(d_2D / 1000) + 0.01*abs(h_UT - h_BS) + 0.83);
				log_ZSD_mean[NLOS] = MAX(-0.5, -3.1*(d_2D / 1000) + 0.01*MAX((h_UT - h_BS), 0) + 0.2);
				log_ZSD_std[LOS] = 0.35;
				log_ZSD_std[NLOS] = 0.35;
				ZOD_offset[LOS] = 0;
				ZOD_offset[NLOS] = -pow(10., -1.5*log10(MAX(10., d_2D)) + 3.3);

			
				ue[ue_idx].ZOD_offset[bs_idx] = ZOD_offset[ue[ue_idx].propagation_condition[bs_idx]];

				if (ue[ue_idx].propagation_condition[bs_idx] == LOS)
				{
					ue[ue_idx].LSPs[bs_idx](LOS_ZSD) = MIN(pow(10., log_ZSD_std[LOS] * ue[ue_idx].TLSPs[bs_idx](LOS_ZSD) + log_ZSD_mean[LOS]),52.);
				}
				else
				{
					ue[ue_idx].LSPs[bs_idx](NLOS_ZSD) = MIN(pow(10., log_ZSD_std[NLOS] * ue[ue_idx].TLSPs[bs_idx](NLOS_ZSD) + log_ZSD_mean[NLOS]),52.);
				}
			}
			else if (pathloss_model == RMa_A || pathloss_model == RMa_A)
			{
				log_ZSD_mean[LOS] = MAX(-1., -0.17*(d_2D / 1000) - 0.01*abs(h_UT - 1.5) + 0.22);
				log_ZSD_mean[NLOS] = MAX(-1., -0.19*(d_2D / 1000) - 0.01*abs(h_UT - 1.5) + 0.28);
				log_ZSD_mean[OUT2IN] = MAX(-1., -0.19*(d_2D / 1000) - 0.01*abs(h_UT - 1.5) + 0.28);
				log_ZSD_std[LOS] = 0.34;
				log_ZSD_std[NLOS] = 0.30;
				log_ZSD_std[OUT2IN] = 0.30;
				ZOD_offset[LOS] = 0;
				ZOD_offset[NLOS] = atan((35. - 3.5) / d_2D) - atan((35. - 1.5) / d_2D);
				ZOD_offset[OUT2IN] = atan((35. - 3.5) / d_2D) - atan((35. - 1.5) / d_2D);


				ue[ue_idx].ZOD_offset[bs_idx] = ZOD_offset[ue[ue_idx].propagation_condition[bs_idx]];

				if (ue[ue_idx].propagation_condition[bs_idx] == LOS)
				{
					ue[ue_idx].LSPs[bs_idx](LOS_ZSD) = MIN(pow(10., log_ZSD_std[LOS] * ue[ue_idx].TLSPs[bs_idx](LOS_ZSD) + log_ZSD_mean[LOS]),52.);
				}
				else if (ue[ue_idx].propagation_condition[bs_idx] == NLOS)
				{
					ue[ue_idx].LSPs[bs_idx](NLOS_ZSD) = MIN(pow(10., log_ZSD_std[NLOS] * ue[ue_idx].TLSPs[bs_idx](NLOS_ZSD) + log_ZSD_mean[NLOS]),52.);
				}
				else
				{
					ue[ue_idx].LSPs[bs_idx](OUT2IN_ZSD) = MIN(pow(10., log_ZSD_std[OUT2IN] * ue[ue_idx].TLSPs[bs_idx](OUT2IN_ZSD) + log_ZSD_mean[OUT2IN]),52.);
				}

			}
			else
			{
				cout << "check pathloss_model" << endl;
			}
		}
	}

}

double h_m(double x)
{
	return exp(-x);
}

void Check_corr_var()
{
	MatrixXd corr_var_LOS = MatrixXd::Zero(7, 7);
	MatrixXd corr_var_NLOS = MatrixXd::Zero(6, 6);
	MatrixXd corr_var_OUT2IN = MatrixXd::Zero(6, 6);
	
	int num_LSP;

	for (int bs_idx = 0; bs_idx < num_bs; bs_idx++)
	{
		for (int floor_idx = 0; floor_idx < num_floor; floor_idx++)
		{
			for (int propagation_condition = 0; propagation_condition < num_propagation_condition; propagation_condition++)
			{
				if (propagation_condition == LOS) { num_LSP = 7; }
				else							  { num_LSP = 6; }
				for (int x = 0; x < size_x_grid; x++)
				{
					for (int y = 0; y < size_y_grid; y++)
					{
						switch (propagation_condition)
						{
						case LOS:
							corr_var_LOS += TLSPs_map[bs_idx][floor_idx][propagation_condition][x][y].matrix()*TLSPs_map[bs_idx][floor_idx][propagation_condition][x][y].matrix().transpose();
							break;
						case NLOS:
							corr_var_NLOS += TLSPs_map[bs_idx][floor_idx][propagation_condition][x][y].matrix()*TLSPs_map[bs_idx][floor_idx][propagation_condition][x][y].matrix().transpose();
							break;
						case OUT2IN:
							corr_var_OUT2IN += TLSPs_map[bs_idx][floor_idx][propagation_condition][x][y].matrix()*TLSPs_map[bs_idx][floor_idx][propagation_condition][x][y].matrix().transpose();
							break;
						default:
							cout << "check propagation_condition " << endl;
						}
					}
				}
			}
		}
	}

	cout << "--------------------cross correlatin at grid --------------------------" << endl;
	cout << "var_LOS  " << endl;
	cout << corr_var_LOS / num_bs/ num_floor/ size_x_grid / size_y_grid << endl;
	cout << endl;
	
	cout << "var_NLOS  " << endl;
	cout << corr_var_NLOS / num_bs / num_floor / size_x_grid / size_y_grid << endl;
	
	cout << endl;
	
	
	cout << "var_OUT2IN " << endl;
	cout << corr_var_OUT2IN / num_bs / num_floor / size_x_grid / size_y_grid << endl;
	cout << endl;
	
	cout << "---------------------------------------------------------------------------"<<endl;

	getchar();




	corr_var_LOS = MatrixXd::Zero(7, 7);
	corr_var_NLOS = MatrixXd::Zero(6, 6);
	corr_var_OUT2IN = MatrixXd::Zero(6, 6);
	double num_LOS = 0;
	double num_NLOS = 0;
	double num_OUT2IN = 0;

	cout << "--------------------ue cross correlatin  ---------------------------------------" << endl;
	for (int bs_idx = 0; bs_idx < num_bs; bs_idx++)
	{
		for (int ue_idx = 0; ue_idx < num_ue; ue_idx++)
		{
			if (ue[ue_idx].propagation_condition[bs_idx] == LOS)
			{
				corr_var_LOS += ue[ue_idx].TLSPs[bs_idx].matrix() * ue[ue_idx].TLSPs[bs_idx].matrix().transpose();
				num_LOS++;
			}
			else if(ue[ue_idx].propagation_condition[bs_idx] == NLOS)
			{
				corr_var_NLOS += ue[ue_idx].TLSPs[bs_idx].matrix() * ue[ue_idx].TLSPs[bs_idx].matrix().transpose();
				num_NLOS++;
			}
			else if (ue[ue_idx].propagation_condition[bs_idx] == OUT2IN)
			{
				corr_var_OUT2IN += ue[ue_idx].TLSPs[bs_idx].matrix() * ue[ue_idx].TLSPs[bs_idx].matrix().transpose();
				num_OUT2IN++;
			}
			else
			{
				cout << "check Check_corr_var(), ue[ue_idx].propagation_condition[bs_idx]" << endl;
			}
		}
	}

	cout << "num_LOS : " << num_LOS << endl;
	cout << "var_LOS  " << endl;
	cout << corr_var_LOS / num_LOS << endl;
	cout << endl;
	cout << "num_NLOS : " << num_NLOS << endl;
	cout << "var_NLOS  " << endl;
	cout << endl;
	cout<< corr_var_NLOS / num_NLOS << endl;
	cout << endl;
	if (num_OUT2IN != 0)
	{
		cout << "var_OUT2IN : " << num_OUT2IN  << endl;
		cout << "var_OUT2IN " << endl;
		cout<< corr_var_OUT2IN / num_OUT2IN << endl;
		cout << endl;
	}

	cout << "---------------------------------------------------------------------------" << endl;

	getchar();

}
